#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a planetarium plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; Query Google Earth data

;; The Elevation API has the following limits in place:
;;
;;     2,500 requests per day.
;;     512 locations per request.
;;     25,000 total locations per day.
;;
;; URLs are restricted to 2048 characters.

(define (monterey-bay)
  ;; This procedure requests the elevations at locations on a 101x101
  ;; gid between 36,-122.5 and 37,-121.5 inclusive -- the coast and
  ;; sea floor of Monterey Bay.  At 11520 elevations per day, the
  ;; 10201 elevations can be fetched in 0.89 days.
  (request-google-elevations "monterey-bay-101x101.txt"
			     (make-latitudes/longitudes
			      (iota 101 36. .01)
			      (iota 101 -122.5 .01))))

(define (jasper-seamount)
  ;; This procedure requests the elevations at locations on a 101x101
  ;; grid between 30.07,-122.67 and 30.57,-122.17 -- centered on the
  ;; Jasper Seamount (30°26.40′N 122°44.40′W).  At 11520 elevations
  ;; per day, the 10201 elevations can be fetched in 0.89 days.
  (request-google-elevations "jasper-seamount-101x101.txt"
			     (make-latitudes/longitudes
			      (iota 101 30.07 .005)
			      (iota 101 -122.67 .005))))

(define (make-latitudes/longitudes lats lngs)
  (append-map! (lambda (lat)
		 (map (lambda (lng)
			(make-latitude/longitude lat lng))
		      lngs))
	       lats))

(define (request-google-elevations log-file locations)
  ;; This procedure requests 8 elevations once a minute -- 1440
  ;; requests and 11520 elevations per day.  It logs its progress to
  ;; LOG-FILE and is restartable, picking up where the log stops.  In
  ;; fact it won't start if there is no log.
  (if (not (file-exists? log-file))
      (error "Could not find log file:" log-file))

  (load-option 'xml)

  (let loop ((locations (remaining-locations locations
					     (read-elevation-data log-file))))
    (if (not (null? locations))
	(let* ((last (if (< 7 (length locations))
			 (list-tail locations 7)
			 '()))
	       (rest (if (pair? last)
			 (let ((rest (cdr last)))
			   (set-cdr! last '())
			   rest)
			 '())))
	  (log-elevations (request-elevations locations) log-file)
	  (sleep-current-thread 60000)
	  (loop rest)))))

(define (request-elevations locations)
  (let ((time (get-universal-time))
	(url (string-append
	      "http://maps.googleapis.com/maps/api/elevation/"
	      "xml?sensor=false&locations="
	      (decorated-string-append
	       "" "%7C" ""
	       (map (lambda (location)
		      (string-append
		       (number->string (latitude location))
		       ","(number->string (longitude location))))
		    locations)))))
    (let* ((port #;(open-input-file "reply.txt")
		 #;(open-input-gfile url)
		 (open-input-string (http-response-body (http-get url '()))))
	   (reply (read-xml port)))
      (let ((root (xml-document-root reply)))
	(if (not (eq? '|ElevationResponse| (xml-element-name root)))
	    (error "Bogus reply:" (xml->string reply)))
	(let ((root-content (xml-element-content root)))
	  ;; First element: status OK
	  (let ((first (skip-strings root-content)))
	    (let ((element (and (pair? first) (car first))))
	      (if (not (and (eq? '|status| (xml-element-name element))
			    (string=? "OK" (get-content element))))
		  (error "Status not OK:" (xml->string reply))))
	    ;; Rest: <result> with <elevation> and <resolution>.
	    (let loop ((next (and (pair? first) (skip-strings (cdr first))))
		       (elevations '()))
	      (if (null? next)
		  (reverse! elevations)
		  (let ((result (car next)))
		    (if (not (eq? '|result| (xml-element-name result)))
			(error "Non-result:" result))
		    (loop (skip-strings (cdr next))
			  (cons (make-elevation
				 (universal-time->global-time-string time)
				 (string-append
				  (get-named-content '|lat| result)
				  ","(get-named-content '|lng| result))
				 (get-named-content '|elevation| result)
				 (get-named-content '|resolution| result))
				elevations)))))))))))

(define (skip-strings content)
  (if (and (pair? content)
	   (string? (car content)))
      (skip-strings (cdr content))
      content))

(define (remaining-locations locations data)
  (do ((data data (cdr data))
       (locations locations (cdr locations)))
      ((or (null? locations)
	   (null? data))
       (if (null? locations)
	   (error "Nothing to do; data collection is complete.")
	   locations))
    (let ((logged (elevation/lat-lng (car data)))
	  (next (car locations)))
      (if (not (and (= (latitude logged) (latitude next))
		    (= (longitude logged) (longitude next))))
	  (error "Log and list mismatched:" logged next)))))

(define (get-named-content name element)
  (get-content (find-element name (xml-element-content element))))

(define (get-content elt)
  (let ((con (xml-element-content elt)))
    (if (and (pair? con)
	     (string? (car con))
	     (null? (cdr con)))
	(car con)
	(error "Unexpected content:" elt))))

(define (find-element name content)
  (or (let loop ((content content))
	(and (pair? content)
	     (let ((elt (car content)))
	       (if (xml-element? elt)
		   (if (eq? name (xml-element-name elt))
		       elt
		       (or (loop (xml-element-content elt))
			   (loop (cdr content))))
		   (loop (cdr content))))))
      (error "Could not find:" name content)))

(define (read-elevation-data log-file)
  (if (file-exists? log-file)
      (map parse-logged-elevation
	   (call-with-input-file log-file read-lines))
      '()))

(define (parse-logged-elevation line)
  (let ((fields (burst-string line #\tab #f)))
    (make-elevation (list-ref fields 0)
		(list-ref fields 1)
		(list-ref fields 2)
		(list-ref fields 3))))

(define (log-elevations elevations log-file)
  (call-with-append-file log-file
    (lambda (port)
      (for-each (lambda (elevation)
		  (write-string (vector-ref elevation 0) port)
		  (write-char #\tab port)
		  (write-string (vector-ref elevation 1) port)
		  (write-char #\tab port)
		  (write-string (vector-ref elevation 2) port)
		  (write-char #\tab port)
		  (write-string (vector-ref elevation 3) port)
		  (newline port))
		elevations))))

(define (read-lines port)
  (let loop ()
    (let ((line (read-line port)))
      (if (eof-object? line)
	  '()
	  (cons line (loop))))))

(define (make-elevation time-string
			lat/lng-string
			elevation-string
			resolution-string)
  (vector time-string lat/lng-string elevation-string resolution-string))

(define (elevation/time elevation)
  (string->universal-time (vector-ref elevation 0)))

(define (elevation/lat-lng elevation)
  (let ((lat-lng (burst-string (vector-ref elevation 1) #\, #f)))
    (if (= 2 (length lat-lng))
	(make-latitude/longitude (string->number (car lat-lng))
				 (string->number (cadr lat-lng)))
	(error "Bogus lat/lng:" elevation))))

(define (elevation/elevation elevation)
  (or (string->number (vector-ref elevation 2))
      (error "Bogus elevation:" elevation)))

(define (elevation/resolution elevation)
  (or (string->number (vector-ref elevation 3))
      (error "Bogus resolution:" elevation)))