#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a planetarium plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;; Formula numbers and unattributed quotes are from Meeus 2009,
;;; specifically the second English edition (1998) "with corrections
;;; as of August 10, 2009" of _Astronomical_Algorithms_ by Jean Meeus.

(declare (reduce-operator (* flo:*)
			  (sin flo:sin) (cos flo:cos)
			  (asin flo:asin) (atan2 flo:atan2)
			  (- flo:- (null-value 0. single) (group left))
			  (+ flo:+ (null-value 0. none) (group right))))

(define (solar-latitude/longitude time)
  ;; The terrestrial latitude/longitude of the mean Sun (a flo:2d) at
  ;; TIME in Julian Ephemeris Days.
  (let* ((T (JD->T time))
	 (lat/long (solar-equatorial-latitude/longitude T))
	 (theta0 (greenwich-mean-sidereal-time-degrees time T)))
    (set-longitude! lat/long (canonical-longitude
			      (flo:- (longitude lat/long) theta0)))
    lat/long))

(define (solar-equatorial-latitude/longitude T)
  ;; The declination and right-ascension of the mean Sun (a flo:2d),
  ;; BOTH IN DEGREES, at T in Julian centuries.
  ;;
  ;; Note that the Sun's latitude, referred to the ecliptic of the
  ;; date, "never exceeds 1.2 arcseconds", so this procedure uses 0.0.

  (ecliptical->equatorial! (make-latitude/longitude 0. (L0 T))
			   (obliquity T)))

(define-integrable (JD->T JD)
  ;; Julian Day to julian centuries since J2000.0.
  (/ (- JD 2451545.0) 36525.))

(define-integrable-operator (L0 T)
  ;; 25.2: "The geometric mean longitude of the Sun, referred to the
  ;; mean equinox of the date." at time T in Julian centuries.
  (canonical-longitude
   (+ 280.46646 (* T (+ 36000.76983 (* 0.0003032 T))))))

#| Unused.

;(define (solar-true-ecliptical-longitude-and-anomaly T)
  ;; "The [Sun's] true geometric longitude referred to the mean
  ;; equinox of the date" "and its true anomaly" (a flo:2d) at T in
  ;; Julian centuries since J2000.0.
  (let ((L0 (L0 T))
	(M (M T)))
    (let ((C (C M T)))
      (let ((sol (+ L0 C))
	    (nu (+ M C)))
	(flo:2d sol nu)))))

;(define-integrable-operator (M T)
  ;; 25.3: "The mean anomaly of the Sun."
  (+ 357.52911 (* T (- 35999.05029 (* 0.0001537 T)))))

;(define-integrable-operator (e T)
  ;; 25.4: "The eccentricity of the Earth's orbit."
  (- 0.016708634 (* T (+ 0.000042037 (* 0.0000001267 T)))))

;(define-integrable-operator (C M T)
  ;; "The Sun's equation of the center."
  (+ (* (- 1.914602 (* T (+ 0.004817 (* 0.000014 T))))
	(sin M))
     (* (- 0.019993 (* 0.000101 T))
	(sin (* 2 M)))
     (* 0.000289
	(sin (* 3 M)))))

;(define-integrable-operator (R e nu)
  ;; 25.5: "The Sun's radius vector, or the distance between the
  ;; centers of the Sun and the Earth, expressed in astronomical
  ;; units" given the Eccentricity of the Earth's orbit E, and its
  ;; true anomaly NU.
  (/ (* 1.000001018 (- 1. (* e e)))
     (+ 1. (* e (cos nu)))))|#

(define (obliquity T)
  ;; 22.2: "The mean obliquity... adopted by the IAU." given T in
  ;; Julian centuries.  Note that "the accuracy of formula 22.2 is not
  ;; satisfactory over a long period of time." but that the error is
  ;; "10minutes over a period of 4000 years".
  (- (dms->degrees 23. 26. 21.448)
     (* T (- (dms->degrees  0.  0. 46.8150)
	     (* T (+ (dms->degrees  0.  0.  0.00059)
		     (* T (- (dms->degrees  0.  0. 0.001813)))))))))

(define (ecliptical->equatorial! lat/long obliquity)
  ;; 13.3-4: "Transformation from ecliptical into equatorial
  ;; coordinates" (all in degrees) with a given obliquity.
  (let ((lambd* (degrees->radians (longitude lat/long)))
	(beta (degrees->radians (latitude lat/long)))
	(epsilon (degrees->radians obliquity)))
    (let ((delta (asin (+ (* (sin beta) (cos epsilon))
			  (* (cos beta) (sin epsilon) (sin lambd*)))))
	  (alpha (atan2 (- (* (sin lambd*) (cos epsilon))
			   (* (tan beta) (sin epsilon)))
			(cos lambd*))))
      (set-latitude! lat/long (radians->degrees delta))
      (set-longitude! lat/long (radians->degrees alpha))))
  lat/long)

(define (greenwich-mean-sidereal-time-degrees time T)
  ;; 12.4: "The mean sidereal time at Greenwich, expressed in
  ;; degrees" at TIME JD and T Julian centuries.
  (canonical-longitude (theta0 time T)))

(define (theta0 time T)
  (+ 280.46061837
     (* 360.98564736629 (- time 2451545.0))
     (* (* T T) (- 0.000387933 (/ T 38710000.)))))

(define (~= n1 n2 epsilon)
  (flo:< (flo:abs (- n2 n1)) epsilon))

(define (round-to-digits N digits)
  (let ((10^digits (expt 10. digits)))
    (/ (round (* N 10^digits))
       10^digits)))

(define (test-greenwich-mean-sidereal-time)
  ;; "Example 12.b -- Find the mean sidereal time at Greenwich on
  ;; 1987 April 10 at 19:21:00 UT."
  (let ((JD (julian-day 1987 4 (+ 10. (/ (+ 19. (/ 21. 60.)) 24.)))))
    (if (not (= JD 2446896.30625))
	(warn "test-greenwich-mean-sidereal-time incorrect JD:"
	      JD 2446896.30625)))
  (let ((T (JD->T 2446896.30625)))
    (if (not (= (round-to-digits T 8) -0.12727430))
	(warn "test-greenwich-mean-sidereal-time incorrect T:"
	      (round-to-digits T 8) T -0.12727430)))
  (let ((theta0+ (theta0 2446896.30625 -0.12727430)))
    (if (not (= (round-to-digits theta0+ 7) -1677831.2621266))
	(warn "test-greenwich-mean-sidereal-time incorrect theta0+:"
	      (round-to-digits theta0+ 7) -1677831.2621266 theta0+)))
  (let ((theta0 (canonical-longitude -1677831.2621266)))
    (if (not (= (round-to-digits theta0 7) 128.7378734))
	(warn "test-greenwich-mean-sidereal-time incorrect theta0:"
	      (round-to-digits theta0 7) 128.7378734 theta0))))

(define (test-solar-latitude/longitude)
  ;; "Example 25.a -- Calculate the Sun's position on 1992 October 13 at 0h TD.
  (let ((JD (julian-day 1992 10 13)))
    (if (not (= JD 2448908.5))
	(warn "test-solar-latitude/longitude incorrect JD:" JD 2448908.5)))
  (let ((T (JD->T 2448908.5)))
    (if (not (= (round-to-digits T 9) -.072183436))
	(warn "test-solar-latitude/longitude incorrect T:"
	      (round-to-digits -0.072183436 9) T)))
  (let ((L0 (L0 -0.072183436)))
    (if (not (= (round-to-digits L0 5) 201.80720))
	(warn ";test-solar-latitude/longitude incorrect L0:"
	      (round-to-digits L0 5) 201.80720 L0))))